---
lastUpdated: true
description: Rage-FW-RPC is an all-in package with asynchronous RPC implementation for RageMP servers in JS/TS
prev: false
next: false
outline: [2, 4]
---

# Rage FW RPC <Badge type="tip" text="^0.3.0" />
is an all-in package with asynchronous RPC implementation for RageMP servers in JS/TS
## Installation
```shell 
npm i @entityseven/rage-fw-rpc
```
```shell 
pnpm i @entityseven/rage-fw-rpc
```
```shell 
yarn add @entityseven/rage-fw-rpc
```

Import installed package and initialize rpc:

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#installation-->

On client-side you have to also specify the browser you want to refer to for events

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#installationClient-->

::: info
Also see [Rpc Config](#rpc-config)
:::

## Motivation
The idea was to create an extensible package, with various features to simplify the development process and provide as much comfort as possible. It should also be using similar architecture as the framework it was specially built for

Inspired by usage of [rage-rpc](https://github.com/micaww/rage-rpc)

## Features
- Type-safe events via [TS generics](https://www.typescriptlang.org/docs/handbook/2/generics.html), avoiding type wrappers
- Built-in logging options for each environment
- Error-safe developer mode for browser
- Calls can receive response from called environments via Promises (browser -> server -> browser, etc.)
- Actual human-readable errors

## Points
Before reading docs you should at least barely be acknowledged about the patterns used and how ``rage-fw-rpc`` is different in usage than community favorite ``rage-rpc``

### Syntax 
Arguments when calling any event must be wrapped in an array. This is done for proper argument typing. Be aware that as of now arguments in ``register`` are SPREAD and not wrapped in array. It is recommended to at least loosely type events, that require passing arrays as arguments to avoid unwanted outcomes.

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#syntax-->

### Asynchronous events
Keep in mind that chaining events this way:
```ts
// eg. client called server
rpc.register('customServerEvent', async (args: string) => {
    const response: string = await rpc.callBrowserAsync('customBrowserEvent', [
        'hello from server',
    ])
    // do something with response
    return 'response from server'
})
```
has two major issues:
1. Asynchronous events are marked as timed out after 10s and are throwing error if no response was received. This can occur when player has high ping or browser delays the response in any way. Means your event chain may not continue in those cases. Your codebase is entirely up to you, but avoid such issues try not to chain events inside one another
2. [``await``](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/await#description) stops function execution. This means any operations in place of ``do something with response`` that do not rely on response are unnecessarily stopped. This can be avoided using [``.then(response => ...)``](https://dev.to/masteringjs/using-then-vs-async-await-in-javascript-2pma) giving function an opportunity to execute side-operations in parallel with promises

### Cross-env communication
Every environment can call any environment and get a response via Promise
```
server <-> client <-> browser AND server <-> browser
```
You can still use ``call<env>`` methods to refer to environment itself
```ts
// server
rpc.callServer(...)
rpc.callServerAsync(...)
// client
rpc.callClient(...)
rpc.callClientAsync(...)
// browser
rpc.callBrowser(...)
rpc.callBrowserAsync(...)
```
It will behave the same way as using ``call/callAsync`` by just being redirected to it. For better code clearance we do recommend using ``call/callAsync`` instead of ``call<env>/call<env>Async`` to also same a tiny bit of computation resources

- (Extra) Due to async-based logic some IDEs or tools like ESlint can soft-warn you for ignoring promises on events you do not want response from. In that cases you can either
```ts twoslash
<!--@include: @/snippets/rage-fw-rpc/0.3.0/declaration.md-->
// ---cut---
await rpc.callServerAsync('event')
// @annotate: await it, but you will have to mark parent function as async

rpc.callServerAsync('event').then()
// @annotate: or use .then() in cases you do not want/unable to mark it as async (eg. React.useEffect)
```

## Docs

### Rpc Config
These are the options you can specify when creating Rpc instance. Options can be omitted at all if you want so. All options only have effect in current context and have to be specified individually on server/client/browser
```ts
interface RpcWrapperConfig {
    forceBrowserDevMode?: boolean // defaults to false
    debugLogs?: boolean // defaults to false
}
```
``forceBrowserDevMode`` - only has effect on browser-side. Fallback for browser to launch without mp context and without errors (e.g. for development in browser). Keep in mind that using this makes browser-side unavailable at all preventing from all operations except logging. Therefore, is recommended to use in pair with ``debugLogs``

``debugLogs`` -  enables logging all exposed methods to available console. 

::: info Console methods used in different environments
Server/browser: ``console.log``

Client: [``mp.console.logInfo``](https://wiki.rage.mp/wiki/Console::logInfo)
:::

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#rpcConfig-->

### register
Registers a callback function for a specified event

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#register-->

### unregister
Unregisters callback function for a specified event

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#unregister-->

### callClient
Calls a synchronous client-side event from server or browser. Won't return any response from client

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#callClient-->

### callClientAsync
Calls a synchronous client-side event from server or browser. Will return a response from client

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#callClientAsync-->

### callServer
Calls a synchronous server-side event from browser or client. Won't return any response from server

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#callServer-->

### callServerAsync
Calls an asynchronous server-side event from browser or client. Will return a response from server

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#callServerAsync-->

### callBrowser
Calls a synchronous server-side event from browser or client. Won't return any response from browser

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#callBrowser-->

### callBrowserAsync
Calls a synchronous server-side event from browser or client. Will return a response from browser

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#callBrowserAsync-->

### call
Calls a synchronous event in current environment. Won't return any response.

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#call-->

::: warning
You cannot call server on itself
:::

### callAsync
Calls an asynchronous event in current environment. Will return a response. You cannot call server on itself

<!--@include: @/snippets/rage-fw-rpc/0.3.0/snippets.md#callAsync-->

::: warning
You cannot call server on itself
:::

## Errors
- When error is thrown you will get a message of such form
```ts
`${rpcData.knownError}\n` + // error message
`Caller: ${rpcData.calledFrom}\n` + // server/client/browser
`Receiver: ${this.environment_}\n` + // server/client/browser
`Event: ${rpcData.eventName}\n` + 
`Additional Info: ${error}` // actual error object, could be more than one 
```
Hopefully this will give you enough information about the event which throws. Error information is definitely subject to change in future
- Keep in mind that event timeouts are throwing with almost no information due to their current implementation (at least your console can show line number)
- Under the hood we use ``JSON.stringify`` to pass data between environments, but in JS there are a few which [cannot be serialized](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/JSON/stringify#description)

### Codes
These should be clear enough themselves, in other cases refer to here
- ``EVENT_NOT_REGISTERED`` - throws in Promise (rejects) in called environment when event is either already unregistered or not registered yet. If you see this its almost always calling an event before registering
- ``UNKNOWN_ENVIRONMENT`` - throws in any environment that is not recognized as server/client/browser in Rage. Unlike ``rage-rpc`` this is not thrown in browser when launched without mp context IF you specify it in browser [Rpc Config](#rpc-config)
- ``NO_BROWSER`` - throws on client if you failed to specify valid browser for it to refer to when calling browser
- ``EVENT_RESPONSE_TIMEOUT`` - throws in Promise (rejects) when failed to receive a response data from called environment. You may not always want to receive it at all, for now it just works like this. Prefer adding ``catch`` on your events
